@@ -1,6 +1,7 @@
+import atexit
+import threading
 from contextlib import contextmanager
-from typing import Any, Optional, Tuple
-import time
+from typing import Any, Dict, Optional, Tuple
 
 import duckdb
 
@@ -9,11 +10,11 @@ from dbt.adapters.base import Credentials
 from dbt.adapters.sql import SQLConnectionManager
 from dbt.contracts.connection import (
     AdapterRequiredConfig,
-    ConnectionState,
     AdapterResponse,
+    Connection,
+    ConnectionState,
 )
 from dbt.logger import GLOBAL_LOGGER as logger
-
 from dataclasses import dataclass
 
 
@@ -23,14 +24,14 @@ class DuckDBCredentials(Credentials):
     schema: str = "main"
     path: str = ":memory:"
 
-    # any extensions we want to install/load (httpfs, json, etc.)
+    # any extensions we want to install and load (httpfs, parquet, etc.)
     extensions: Optional[Tuple[str, ...]] = None
 
-    # for connecting to data in S3 via the httpfs extension
-    s3_region: Optional[str] = None
-    s3_access_key_id: Optional[str] = None
-    s3_secret_access_key: Optional[str] = None
-    s3_session_token: Optional[str] = None
+    # any additional pragmas we want to configure on our DuckDB connections;
+    # a list of the built-in pragmas can be found here:
+    # https://duckdb.org/docs/sql/configuration
+    # (and extensions may add their own pragmas as well)
+    settings: Optional[Dict[str, Any]] = None
 
     @property
     def type(self):
@@ -62,25 +63,14 @@ class DuckDBConnectionWrapper:
     def __init__(self, conn, credentials):
         self._conn = conn
 
-        # Extensions need to be config'd per cursor for reasons
+        # Extensions/settings need to be configured per cursor
         cursor = conn.cursor()
         for ext in credentials.extensions or []:
             cursor.execute(f"LOAD '{ext}'")
-        if credentials.s3_region is not None:
-            cursor.execute("LOAD 'httpfs'")
-            cursor.execute(f"SET s3_region = '{credentials.s3_region}'")
-            if credentials.s3_access_key_id is not None:
-                cursor.execute(
-                    f"SET s3_access_key_id = '{credentials.s3_access_key_id}'"
-                )
-                cursor.execute(
-                    f"SET s3_secret_access_key = '{credentials.s3_secret_access_key}'"
-                )
-            else:
-                cursor.execute(
-                    f"SET s3_session_token = '{credentials.s3_session_token}'"
-                )
-
+        for key, value in (credentials.settings or {}).items():
+            # Okay to set these as strings because DuckDB will cast them
+            # to the correct type
+            cursor.execute(f"SET {key} = '{value}'")
         self._cursor = DuckDBCursorWrapper(cursor)
 
     def __getattr__(self, name):
@@ -92,52 +82,60 @@ class DuckDBConnectionWrapper:
 
 class DuckDBConnectionManager(SQLConnectionManager):
     TYPE = "duckdb"
+    LOCK = threading.RLock()
+    CONN = None
+    CONN_COUNT = 0
 
     def __init__(self, profile: AdapterRequiredConfig):
         super().__init__(profile)
-        if profile.threads > 1:
-            raise dbt.exceptions.RuntimeException(
-                "dbt-duckdb only supports 1 thread at this time"
-            )
 
     @classmethod
-    def open(cls, connection):
+    def open(cls, connection: Connection) -> Connection:
         if connection.state == ConnectionState.OPEN:
             logger.debug("Connection is already open, skipping open.")
             return connection
 
         credentials = cls.get_credentials(connection.credentials)
-        try:
-            conn = duckdb.connect(credentials.path, read_only=False)
+        with cls.LOCK:
+            try:
+                if not cls.CONN:
+                    cls.CONN = duckdb.connect(credentials.path, read_only=False)
 
-            # install any extensions on the connection
-            if credentials.extensions is not None:
-                for extension in credentials.extensions:
-                    conn.execute(f"INSTALL '{extension}'")
+                    # install any extensions on the connection
+                    if credentials.extensions is not None:
+                        for extension in credentials.extensions:
+                            cls.CONN.execute(f"INSTALL '{extension}'")
 
-            if credentials.s3_region is not None:
-                conn.execute("INSTALL 'httpfs'")
-                if credentials.s3_session_token is None and (
-                    credentials.s3_access_key_id is None
-                    or credentials.s3_secret_access_key is None
-                ):
-                    raise dbt.exceptions.RuntimeException(
-                        "You must specify either s3_session_token or s3_access_key_id and s3_secret_access_key"
-                    )
+                connection.handle = DuckDBConnectionWrapper(cls.CONN.cursor(), credentials)
+                connection.state = ConnectionState.OPEN
+                cls.CONN_COUNT += 1
 
-            connection.handle = DuckDBConnectionWrapper(conn, credentials)
-            connection.state = ConnectionState.OPEN
+            except RuntimeError as e:
+                logger.debug(
+                    "Got an error when attempting to open a duckdb " "database: '{}'".format(e)
+                )
 
-        except RuntimeError as e:
-            logger.debug(
-                "Got an error when attempting to open a duckdb "
-                "database: '{}'".format(e)
-            )
+                connection.handle = None
+                connection.state = ConnectionState.FAIL
 
-            connection.handle = None
-            connection.state = ConnectionState.FAIL
+                raise dbt.exceptions.FailedToConnectException(str(e))
+            return connection
 
-            raise dbt.exceptions.FailedToConnectException(str(e))
+    @classmethod
+    def close(cls, connection: Connection) -> Connection:
+        # if the connection is in closed or init, there's nothing to do
+        if connection.state in {ConnectionState.CLOSED, ConnectionState.INIT}:
+            return connection
+
+        connection = super(SQLConnectionManager, cls).close(connection)
+
+        if connection.state == ConnectionState.CLOSED:
+            credentials = cls.get_credentials(connection.credentials)
+            with cls.LOCK:
+                cls.CONN_COUNT -= 1
+                if cls.CONN_COUNT == 0 and cls.CONN and not credentials.path == ":memory:":
+                    cls.CONN.close()
+                    cls.CONN = None
 
         return connection
 
@@ -148,7 +146,7 @@ class DuckDBConnectionManager(SQLConnectionManager):
     def exception_handler(self, sql: str, connection_name="master"):
         try:
             yield
-        except dbt.exceptions.RuntimeException as dbte:
+        except dbt.exceptions.RuntimeException:
             raise
         except RuntimeError as e:
             logger.debug("duckdb error: {}".format(str(e)))
@@ -166,3 +164,13 @@ class DuckDBConnectionManager(SQLConnectionManager):
         # https://github.com/dbt-labs/dbt-spark/issues/142
         message = "OK"
         return AdapterResponse(_message=message)
+
+    @classmethod
+    def close_all_connections(cls):
+        with cls.LOCK:
+            if cls.CONN is not None:
+                cls.CONN.close()
+                cls.CONN = None
+
+
+atexit.register(DuckDBConnectionManager.close_all_connections)
